@@ -5,6 +5,7 @@ from .gemini_provider import GeminiProvider
 from .together_provider import TogetherProvider
 from .deepseek_provider import DeepSeekProvider
 from .mistral_provider import MistralProvider
+from .grok_provider import GrokProvider
 
 __all__ = [
     "BaseLLMProvider",
@@ -15,4 +16,5 @@ __all__ = [
     "TogetherProvider",
     "DeepSeekProvider",
     "MistralProvider",
+    "GrokProvider",
 ]
