@@ -20,25 +20,30 @@ logger = logging.getLogger(__name__)
 class AnthropicProvider(BaseLLMProvider):
     """Anthropic Claude provider implementation."""
 
-    def __init__(self, api_key: Optional[str] = None, config=None):
+    def __init__(
+        self,
+        api_key: Optional[str] = None,
+        base_url: Optional[str] = None,
+        config=None,
+    ):
         super().__init__(
-            api_key or defog_config.get("ANTHROPIC_API_KEY"), config=config
+            api_key or defog_config.get("ANTHROPIC_API_KEY"),
+            base_url=base_url,
+            config=config,
         )
 
     @classmethod
     def from_config(cls, config: LLMConfig):
         """Create Anthropic provider from config."""
-        return cls(api_key=config.get_api_key("anthropic"), config=config)
+        return cls(
+            api_key=config.get_api_key("anthropic"),
+            base_url=config.get_base_url("anthropic"),
+            config=config,
+        )
 
     def get_provider_name(self) -> str:
         return "anthropic"
 
-    def supports_tools(self, model: str) -> bool:
-        return True  # All current Claude models support tools
-
-    def supports_response_format(self, model: str) -> bool:
-        return True  # All current Claude models support structured output via system prompts
-
     def convert_content_to_anthropic(self, content: Any) -> Any:
         """Convert message content to Anthropic format."""
         return convert_to_anthropic_format(content)
@@ -109,7 +114,6 @@ class AnthropicProvider(BaseLLMProvider):
         max_completion_tokens: Optional[int] = None,
         temperature: float = 0.0,
         response_format=None,
-        seed: int = 0,
         tools: Optional[List[Callable]] = None,
         tool_choice: Optional[str] = None,
         store: bool = True,
@@ -117,7 +121,7 @@ class AnthropicProvider(BaseLLMProvider):
         timeout: int = 600,
         prediction: Optional[Dict[str, str]] = None,
         reasoning_effort: Optional[str] = None,
-        mcp_servers: Optional[List[Dict[str, Any]]] = None,
+        parallel_tool_calls: bool = False,
         **kwargs,
     ) -> Tuple[Dict[str, Any], List[Dict[str, Any]]]:
         """Create the parameter dict for Anthropic's .messages.create()."""
@@ -185,7 +189,8 @@ class AnthropicProvider(BaseLLMProvider):
         }
 
         # Handle structured output for Anthropic models
-        if response_format:
+        # When tools are provided, we'll set response_format later in the final call
+        if response_format and not (tools and len(tools) > 0):
             # Add instructions to the latest user message to enforce structured output
             if isinstance(response_format, type) and hasattr(
                 response_format, "model_json_schema"
@@ -235,14 +240,14 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
                 params["tool_choice"] = {"type": "auto"}
 
             # Add parallel tool calls configuration
-            if "tool_choice" in params and isinstance(params["tool_choice"], dict):
-                if not self.config.enable_parallel_tool_calls:
+            if (
+                "tool_choice" in params
+                and isinstance(params["tool_choice"], dict)
+                and not model.startswith("grok")
+            ):
+                if not parallel_tool_calls:
                     params["tool_choice"]["disable_parallel_tool_use"] = True
 
-        # Add MCP servers if provided
-        if mcp_servers:
-            params["mcp_servers"] = mcp_servers
-
         return params, messages
 
     async def process_response(
@@ -254,9 +259,9 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
         tool_dict: Dict[str, Callable],
         response_format=None,
         post_tool_function: Optional[Callable] = None,
-        mcp_servers: Optional[List[Dict[str, Any]]] = None,
-        image_result_keys: Optional[List[str]] = None,
+        post_response_hook: Optional[Callable] = None,
         tool_handler: Optional[ToolHandler] = None,
+        return_tool_outputs_only: bool = False,
         **kwargs,
     ) -> Tuple[
         Any, List[Dict[str, Any]], int, int, Optional[int], Optional[Dict[str, int]]
@@ -281,9 +286,13 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
         tool_outputs = []
         total_input_tokens = 0
         total_output_tokens = 0
+        return_tool_outputs_only = bool(return_tool_outputs_only)
+
+        def has_tool_call_outputs() -> bool:
+            return any(output.get("tool_call_id") for output in tool_outputs)
 
         # Handle tool processing for both local tools and MCP server tools
-        if (tools and len(tools) > 0) or (mcp_servers and len(mcp_servers) > 0):
+        if tools and len(tools) > 0:
             consecutive_exceptions = 0
             while True:
                 total_input_tokens += response.usage.input_tokens
@@ -313,6 +322,15 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
                     for block in response.content
                     if hasattr(block, "type") and block.type == "text"
                 ]
+
+                # call this at the start of the while loop
+                # to ensure we also log the first message (that comes in the function arg)
+                await self.call_post_response_hook(
+                    post_response_hook=post_response_hook,
+                    response=response,
+                    messages=request_params.get("messages", []),
+                )
+
                 if len(tool_call_blocks) > 0:
                     try:
                         # Separate MCP tools from regular tools
@@ -369,6 +387,9 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
                                 post_tool_function,
                                 consecutive_exceptions,
                                 tool_handler=tool_handler,
+                                parallel_tool_calls=kwargs.get(
+                                    "parallel_tool_calls", True
+                                ),
                             )
 
                         # For MCP tools, extract results from mcp_tool_result blocks
@@ -464,7 +485,9 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
 
                                 # Build user response with all tool results and handle images
                                 tool_results_data = process_tool_results_with_images(
-                                    tool_call_blocks, results, image_result_keys
+                                    tool_call_blocks,
+                                    results,
+                                    tool_handler.image_result_keys,
                                 )
 
                                 # Build tool results content
@@ -583,16 +606,104 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
                         tools, tool_handler, request_params, request_params.get("model")
                     )
 
-                    # Make next call - use the same API endpoint (beta or regular) as the initial call
-                    if mcp_servers and len(mcp_servers) > 0:
-                        response = await client.beta.messages.create(**request_params)
-                    else:
-                        response = await client.messages.create(**request_params)
+                    response = await client.messages.create(**request_params)
                 else:
                     # Break out of loop when tool calls are finished
+                    skip_final_response = (
+                        return_tool_outputs_only and has_tool_call_outputs()
+                    )
+                    if skip_final_response:
+                        content = ""
+                        break
+
                     content = "\n".join([block.text for block in text_blocks])
+
+                    # If we have response_format and tools were being used, make one more call
+                    if response_format and request_params.get("tools"):
+                        # Add the assistant's response to messages
+                        assistant_content = []
+                        if len(thinking_blocks) > 0:
+                            assistant_content += thinking_blocks
+                        for block in text_blocks:
+                            assistant_content.append(block)
+
+                        request_params["messages"].append(
+                            {
+                                "role": "assistant",
+                                "content": assistant_content,
+                            }
+                        )
+
+                        # Remove tools and add structured output instructions
+                        request_params.pop("tools", None)
+                        request_params.pop("tool_choice", None)
+
+                        # Re-add structured output instructions to the conversation
+                        if isinstance(response_format, type) and hasattr(
+                            response_format, "model_json_schema"
+                        ):
+                            schema = response_format.model_json_schema()
+                            schema_str = json.dumps(schema, indent=2)
+
+                            structured_instruction = f"""
+
+IMPORTANT: You must respond with ONLY a valid, properly formatted JSON object that conforms to the following JSON schema:
+{schema_str}
+
+RESPONSE FORMAT INSTRUCTIONS:
+1. Your entire response must be ONLY the JSON object, with no additional text before or after.
+2. Format the JSON properly with no line breaks within property values.
+3. Use double quotes for all property names and string values.
+4. Do not add comments or explanations outside the JSON structure.
+5. Ensure all required properties in the schema are included.
+6. Make sure the JSON is properly formatted and can be parsed by standard JSON parsers.
+
+THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS BEFORE OR AFTER.
+"""
+                            # Add a new user message with the structured output request
+                            request_params["messages"].append(
+                                {
+                                    "role": "user",
+                                    "content": [
+                                        {
+                                            "type": "text",
+                                            "text": "Based on the tool results above, please provide the final answer in the requested JSON format."
+                                            + structured_instruction,
+                                        }
+                                    ],
+                                }
+                            )
+
+                        # Make final call for structured output
+                        response = await client.messages.create(**request_params)
+
+                        await self.call_post_response_hook(
+                            post_response_hook=post_response_hook,
+                            response=response,
+                            messages=request_params.get("messages", []),
+                        )
+
+                        # Extract final content
+                        content = ""
+                        for block in response.content:
+                            if hasattr(block, "type") and block.type == "text":
+                                content = block.text
+                                break
+
+                        # Update token counts
+                        total_input_tokens += (
+                            response.usage.input_tokens
+                            + response.usage.cache_creation_input_tokens
+                        )
+                        total_output_tokens += response.usage.output_tokens
+
                     break
         else:
+            await self.call_post_response_hook(
+                post_response_hook=post_response_hook,
+                response=response,
+                messages=request_params.get("messages", []),
+            )
             # No tools provided
             content = ""
             for block in response.content:
@@ -600,8 +711,11 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
                     content = block.text
                     break
 
+        if return_tool_outputs_only and has_tool_call_outputs():
+            content = ""
+
         # Parse structured output if response_format is provided
-        if response_format and not tools:
+        if response_format:
             # Use base class method for structured response parsing
             content = self.parse_structured_response(content, response_format)
 
@@ -626,7 +740,6 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
         max_completion_tokens: Optional[int] = None,
         temperature: float = 0.0,
         response_format=None,
-        seed: int = 0,
         tools: Optional[List[Callable]] = None,
         tool_choice: Optional[str] = None,
         store: bool = True,
@@ -635,16 +748,20 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
         prediction: Optional[Dict[str, str]] = None,
         reasoning_effort: Optional[str] = None,
         post_tool_function: Optional[Callable] = None,
-        mcp_servers: Optional[List[Dict[str, Any]]] = None,
+        post_response_hook: Optional[Callable] = None,
         image_result_keys: Optional[List[str]] = None,
         tool_budget: Optional[Dict[str, int]] = None,
+        previous_response_id: Optional[str] = None,
         **kwargs,
     ) -> LLMResponse:
         """Execute a chat completion with Anthropic."""
         from anthropic import AsyncAnthropic
 
-        # Create a ToolHandler instance with tool_budget if provided
-        tool_handler = self.create_tool_handler_with_budget(tool_budget)
+        # Create a ToolHandler instance with tool_budget and image_result_keys if provided
+        tool_handler = self.create_tool_handler_with_budget(
+            tool_budget, image_result_keys, kwargs.get("tool_output_max_tokens")
+        )
+        return_tool_outputs_only = kwargs.pop("return_tool_outputs_only", False)
 
         if post_tool_function:
             tool_handler.validate_post_tool_function(post_tool_function)
@@ -653,21 +770,27 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
 
         # Set up headers based on whether MCP servers are provided
         headers = {}
-        if mcp_servers:
-            headers["anthropic-beta"] = "mcp-client-2025-04-04"
-        else:
-            headers["anthropic-beta"] = "interleaved-thinking-2025-05-14"
+        headers["anthropic-beta"] = "interleaved-thinking-2025-05-14"
 
-        client = AsyncAnthropic(
-            api_key=self.api_key,
-            default_headers=headers,
-        )
+        client_kwargs = {
+            "api_key": self.api_key,
+            "default_headers": headers,
+        }
+
+        if self.base_url:
+            client_kwargs["base_url"] = self.base_url
+
+        client = AsyncAnthropic(**client_kwargs)
 
         # Filter tools based on budget before building params
         tools = self.filter_tools_by_budget(tools, tool_handler)
 
+        conversation_messages = self.prepare_conversation_messages(
+            messages, previous_response_id
+        )
+
         params, _ = self.build_params(
-            messages=messages,
+            messages=conversation_messages,
             model=model,
             max_completion_tokens=max_completion_tokens,
             temperature=temperature,
@@ -676,7 +799,7 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
             response_format=response_format,
             reasoning_effort=reasoning_effort,
             timeout=timeout,
-            mcp_servers=mcp_servers,
+            parallel_tool_calls=kwargs.get("parallel_tool_calls", True),
         )
 
         # Construct a tool dict if needed
@@ -684,10 +807,7 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
         if tools and len(tools) > 0 and "tools" in params:
             tool_dict = tool_handler.build_tool_dict(tools)
 
-        if mcp_servers and len(mcp_servers) > 0:
-            func_to_call = client.beta.messages.create
-        else:
-            func_to_call = client.messages.create
+        func_to_call = client.messages.create
 
         try:
             response = await func_to_call(**params)
@@ -707,14 +827,25 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
                 tool_dict=tool_dict,
                 response_format=response_format,
                 post_tool_function=post_tool_function,
-                mcp_servers=mcp_servers,
-                image_result_keys=image_result_keys,
+                post_response_hook=post_response_hook,
                 tool_handler=tool_handler,
+                return_tool_outputs_only=return_tool_outputs_only,
+                **kwargs,
             )
         except Exception as e:
             traceback.print_exc()
             raise ProviderError(self.get_provider_name(), f"API call failed: {e}", e)
 
+        api_response_id = getattr(response, "id", None)
+        response_id = api_response_id
+        if store:
+            cache_response_id = api_response_id or self.generate_response_id()
+            history_for_cache = self.append_assistant_message_to_history(
+                conversation_messages, content
+            )
+            self.persist_conversation_history(cache_response_id, history_for_cache)
+            response_id = cache_response_id
+
         # Calculate cost
         cost = CostCalculator.calculate_cost(
             model, input_toks, output_toks, cached_toks
@@ -730,4 +861,5 @@ THE RESPONSE SHOULD START WITH '{{' AND END WITH '}}' WITH NO OTHER CHARACTERS B
             output_tokens_details=output_details,
             cost_in_cents=cost,
             tool_outputs=tool_outputs,
+            response_id=response_id,
         )
