@@ -39,14 +39,6 @@ class DeepSeekProvider(BaseLLMProvider):
     def get_provider_name(self) -> str:
         return "deepseek"
 
-    def supports_tools(self, model: str) -> bool:
-        # Only deepseek-chat supports tools, deepseek-reasoner does not
-        return model == "deepseek-chat"
-
-    def supports_response_format(self, model: str) -> bool:
-        # Both models support JSON response format
-        return True
-
     def _get_media_type(self, img_data: str) -> str:
         """Detect media type from base64 image data."""
         try:
@@ -111,7 +103,6 @@ class DeepSeekProvider(BaseLLMProvider):
         max_completion_tokens: Optional[int] = None,
         temperature: float = 0.0,
         response_format=None,
-        seed: int = 0,
         tools: Optional[List[Callable]] = None,
         tool_choice: Optional[str] = None,
         store: bool = True,
@@ -119,7 +110,6 @@ class DeepSeekProvider(BaseLLMProvider):
         timeout: int = 600,
         prediction: Optional[Dict[str, str]] = None,
         reasoning_effort: Optional[str] = None,
-        mcp_servers: Optional[List[Dict[str, Any]]] = None,
         **kwargs,
     ) -> Tuple[Dict[str, Any], List[Dict[str, str]]]:
         """
@@ -132,7 +122,6 @@ class DeepSeekProvider(BaseLLMProvider):
             "messages": messages,
             "model": model,
             "max_completion_tokens": max_completion_tokens,
-            "seed": seed,
             "store": store,
             "metadata": metadata,
             "timeout": timeout,
@@ -144,12 +133,6 @@ class DeepSeekProvider(BaseLLMProvider):
 
         # Tools are supported for deepseek-chat
         if tools and len(tools) > 0:
-            if not self.supports_tools(model):
-                raise ProviderError(
-                    self.get_provider_name(),
-                    f"Model {model} does not support tools/function calling",
-                )
-
             function_specs = get_function_specs(tools, model)
             request_params["tools"] = function_specs
             if tool_choice:
@@ -165,7 +148,7 @@ class DeepSeekProvider(BaseLLMProvider):
             )
 
         # Handle structured output for DeepSeek models
-        if response_format and self.supports_response_format(model):
+        if response_format:
             # Check if response_format is a Pydantic model
             if isinstance(response_format, type) and hasattr(
                 response_format, "model_json_schema"
@@ -203,12 +186,6 @@ Respond with JSON only.
                 # For non-Pydantic response formats, use the standard approach
                 request_params["response_format"] = response_format
 
-        # DeepSeek doesn't support MCP servers
-        if mcp_servers:
-            raise ProviderError(
-                self.get_provider_name(), "DeepSeek does not support MCP servers"
-            )
-
         return request_params, messages
 
     async def process_response(
@@ -406,7 +383,6 @@ Respond with JSON only.
         max_completion_tokens: Optional[int] = None,
         temperature: float = 0.0,
         response_format=None,
-        seed: int = 0,
         tools: Optional[List[Callable]] = None,
         tool_choice: Optional[str] = None,
         store: bool = True,
@@ -415,7 +391,6 @@ Respond with JSON only.
         prediction: Optional[Dict[str, str]] = None,
         reasoning_effort: Optional[str] = None,
         post_tool_function: Optional[Callable] = None,
-        mcp_servers: Optional[List[Dict[str, Any]]] = None,
         tool_budget: Optional[Dict[str, int]] = None,
         **kwargs,
     ) -> LLMResponse:
@@ -423,7 +398,9 @@ Respond with JSON only.
         from openai import AsyncOpenAI
 
         # Create a ToolHandler instance with tool_budget if provided
-        tool_handler = self.create_tool_handler_with_budget(tool_budget)
+        tool_handler = self.create_tool_handler_with_budget(
+            tool_budget, None, kwargs.get("tool_output_max_tokens")
+        )
 
         if post_tool_function:
             tool_handler.validate_post_tool_function(post_tool_function)
@@ -440,7 +417,6 @@ Respond with JSON only.
             max_completion_tokens=max_completion_tokens,
             temperature=temperature,
             response_format=response_format,
-            seed=seed,
             tools=tools,
             tool_choice=tool_choice,
             prediction=prediction,
@@ -448,7 +424,6 @@ Respond with JSON only.
             store=store,
             metadata=metadata,
             timeout=timeout,
-            mcp_servers=mcp_servers,
         )
 
         # Build a tool dict if needed
