@@ -1,10 +1,12 @@
-from defog import config as defog_config
 import traceback
 import time
 import base64
 import logging
+import uuid
 from typing import Dict, List, Any, Optional, Callable, Tuple, Union
 
+from defog import config as defog_config
+
 from google import genai
 from google.genai.types import (
     Part,
@@ -42,6 +44,21 @@ class GeminiProvider(BaseLLMProvider):
     def get_provider_name(self) -> str:
         return "gemini"
 
+    def _get_or_assign_tool_id(self, tool_call: Any) -> str:
+        """Ensure tool call has a stable id even when Gemini omits it."""
+        tool_id = getattr(tool_call, "id", None)
+        if tool_id:
+            return tool_id
+
+        tool_id = f"gemini-tool-{uuid.uuid4().hex}"
+        try:
+            setattr(tool_call, "id", tool_id)
+        except Exception:
+            logger.debug(
+                "Unable to assign generated tool_id to Gemini function call instance"
+            )
+        return tool_id
+
     def _get_media_type(self, img_data: str) -> str:
         """Detect media type from base64 image data."""
         try:
@@ -113,12 +130,6 @@ class GeminiProvider(BaseLLMProvider):
 
         return Content(role="user", parts=parts)
 
-    def supports_tools(self, model: str) -> bool:
-        return True  # All current Gemini models support tools
-
-    def supports_response_format(self, model: str) -> bool:
-        return True  # All current Gemini models support structured output
-
     def convert_content_to_gemini_parts(self, content: Any, genai_types) -> List[Any]:
         """Convert message content to Gemini Part objects."""
         return convert_to_gemini_parts(content, genai_types)
@@ -130,7 +141,6 @@ class GeminiProvider(BaseLLMProvider):
         max_completion_tokens: Optional[int] = None,
         temperature: float = 0.0,
         response_format: Optional[Any] = None,
-        seed: int = 0,
         tools: Optional[List[Callable]] = None,
         tool_choice: Optional[str] = None,
         store: bool = True,
@@ -138,7 +148,7 @@ class GeminiProvider(BaseLLMProvider):
         timeout: int = 600,
         prediction: Optional[Dict[str, str]] = None,
         reasoning_effort: Optional[str] = None,
-        mcp_servers: Optional[List[Dict[str, Any]]] = None,
+        parallel_tool_calls: bool = False,
         **kwargs,
     ) -> Tuple[Dict[str, Any], List[Dict[str, Any]]]:
         """Construct parameters for Gemini's generate_content call."""
@@ -219,7 +229,8 @@ class GeminiProvider(BaseLLMProvider):
             # The model decides when to call multiple functions in parallel
             # This is controlled internally and cannot be disabled
 
-        if response_format:
+        # When tools are provided, we'll set response_format later in the final call
+        if response_format and not (tools and len(tools) > 0):
             # If we want a JSON / Pydantic format
             # "response_schema" is only recognized if the google.genai library supports it
             request_params["response_mime_type"] = "application/json"
@@ -238,8 +249,9 @@ class GeminiProvider(BaseLLMProvider):
         response_format: Optional[Any] = None,
         model: str = "",
         post_tool_function: Optional[Callable] = None,
-        image_result_keys: Optional[List[str]] = None,
+        post_response_hook: Optional[Callable] = None,
         tool_handler: Optional[ToolHandler] = None,
+        return_tool_outputs_only: bool = False,
         **kwargs,
     ) -> Tuple[
         Any, List[Dict[str, Any]], int, int, Optional[int], Optional[Dict[str, int]]
@@ -260,6 +272,11 @@ class GeminiProvider(BaseLLMProvider):
         tool_outputs = []
         total_input_tokens = 0
         total_output_tokens = 0
+        return_tool_outputs_only = bool(return_tool_outputs_only)
+
+        def has_tool_outputs() -> bool:
+            return any(output.get("tool_call_id") for output in tool_outputs)
+
         if tools and len(tools) > 0:
             consecutive_exceptions = 0
             while True:
@@ -270,6 +287,15 @@ class GeminiProvider(BaseLLMProvider):
                 total_output_tokens += (
                     response.usage_metadata.candidates_token_count or 0
                 )
+
+                # call this at the start of the while loop
+                # to ensure we also log the first message (that comes in the function arg)
+                await self.call_post_response_hook(
+                    post_response_hook=post_response_hook,
+                    response=response,
+                    messages=request_params.get("messages", []),
+                )
+
                 if response.function_calls:
                     try:
                         # Prepare tool calls for batch execution
@@ -277,8 +303,7 @@ class GeminiProvider(BaseLLMProvider):
                         for tool_call in response.function_calls:
                             func_name = tool_call.name
                             args = tool_call.args
-                            # set tool_id to None, as Gemini models do not return a tool_id by default
-                            tool_id = getattr(tool_call, "id", None)
+                            tool_id = self._get_or_assign_tool_id(tool_call)
 
                             tool_calls_batch.append(
                                 {
@@ -318,10 +343,11 @@ class GeminiProvider(BaseLLMProvider):
                         for tool_call, result in zip(response.function_calls, results):
                             func_name = tool_call.name
                             args = tool_call.args
-                            tool_id = getattr(tool_call, "id", None)
+                            tool_id = self._get_or_assign_tool_id(tool_call)
 
                             tool_outputs.append(
                                 {
+                                    "tool_call_id": tool_id,
                                     "tool_id": tool_id,
                                     "name": func_name,
                                     "args": args,
@@ -338,7 +364,9 @@ class GeminiProvider(BaseLLMProvider):
                         # print(results, image_result_keys)
 
                         tool_data_list = process_tool_results_with_images(
-                            response.function_calls, results, image_result_keys
+                            response.function_calls,
+                            results,
+                            tool_handler.image_result_keys,
                         )
 
                         # Create Gemini-specific messages
@@ -383,6 +411,11 @@ class GeminiProvider(BaseLLMProvider):
                         request_params["tool_config"] = ToolConfig(
                             function_calling_config=FunctionCallingConfig(mode="AUTO")
                         )
+
+                        # If no more tools are available and we have response_format, set it now
+                        if not tools and response_format:
+                            request_params["response_mime_type"] = "application/json"
+                            request_params["response_schema"] = response_format
                     except ProviderError:
                         # Re-raise provider errors from base class
                         raise
@@ -416,15 +449,63 @@ class GeminiProvider(BaseLLMProvider):
                     )
                 else:
                     # Break out of loop when tool calls are finished
+                    if return_tool_outputs_only and has_tool_outputs():
+                        content = ""
+                        break
+
                     content = response.text.strip() if response.text else None
+
+                    # If we have response_format and tools were being used, make one more call
+                    if response_format and request_params.get("tools"):
+                        # Add the assistant's response to messages
+                        messages.append(
+                            Content(
+                                role="model",
+                                parts=[Part.from_text(text=content)] if content else [],
+                            )
+                        )
+
+                        # Remove tools and set response format
+                        request_params.pop("tools", None)
+                        request_params.pop("tool_config", None)
+                        request_params.pop("automatic_function_calling", None)
+                        request_params["response_mime_type"] = "application/json"
+                        request_params["response_schema"] = response_format
+
+                        # Make final call for structured output
+                        response = await client.aio.models.generate_content(
+                            model=model,
+                            contents=messages,
+                            config=GenerateContentConfig(**request_params),
+                        )
+
+                        await self.call_post_response_hook(
+                            post_response_hook=post_response_hook,
+                            response=response,
+                            messages=request_params.get("messages", []),
+                        )
+
+                        content = response.text.strip() if response.text else None
                     break
         else:
+            # call this at the start of the while loop
+            # to ensure we also log the first message (that comes in the function arg)
+            await self.call_post_response_hook(
+                post_response_hook=post_response_hook,
+                response=response,
+                messages=request_params.get("messages", []),
+            )
+
             # No tools provided
-            if response_format:
-                # Use base class method for structured response parsing
-                content = self.parse_structured_response(response.text, response_format)
-            else:
-                content = response.text.strip() if response.text else None
+            content = response.text.strip() if response.text else None
+
+        if return_tool_outputs_only and has_tool_outputs():
+            content = ""
+
+        # Parse structured output if response_format is provided
+        if response_format and content:
+            # Use base class method for structured response parsing
+            content = self.parse_structured_response(content, response_format)
 
         usage = response.usage_metadata
         total_input_tokens += usage.prompt_token_count or 0
@@ -445,7 +526,6 @@ class GeminiProvider(BaseLLMProvider):
         max_completion_tokens: Optional[int] = None,
         temperature: float = 0.0,
         response_format: Optional[Any] = None,
-        seed: int = 0,
         tools: Optional[List[Callable]] = None,
         tool_choice: Optional[str] = None,
         store: bool = True,
@@ -454,14 +534,18 @@ class GeminiProvider(BaseLLMProvider):
         prediction: Optional[Dict[str, str]] = None,
         reasoning_effort: Optional[str] = None,
         post_tool_function: Optional[Callable] = None,
-        mcp_servers: Optional[List[Dict[str, Any]]] = None,
+        post_response_hook: Optional[Callable] = None,
         image_result_keys: Optional[List[str]] = None,
         tool_budget: Optional[Dict[str, int]] = None,
+        previous_response_id: Optional[str] = None,
         **kwargs,
     ) -> LLMResponse:
         """Execute a chat completion with Gemini."""
-        # Create a ToolHandler instance with tool_budget if provided
-        tool_handler = self.create_tool_handler_with_budget(tool_budget)
+        # Create a ToolHandler instance with tool_budget and image_result_keys if provided
+        tool_handler = self.create_tool_handler_with_budget(
+            tool_budget, image_result_keys, kwargs.get("tool_output_max_tokens")
+        )
+        return_tool_outputs_only = kwargs.pop("return_tool_outputs_only", False)
 
         if post_tool_function:
             tool_handler.validate_post_tool_function(post_tool_function)
@@ -472,13 +556,16 @@ class GeminiProvider(BaseLLMProvider):
         # Filter tools based on budget before building params
         tools = self.filter_tools_by_budget(tools, tool_handler)
 
-        request_params, messages = self.build_params(
-            messages=messages,
+        conversation_messages = self.prepare_conversation_messages(
+            messages, previous_response_id
+        )
+
+        request_params, provider_messages = self.build_params(
+            messages=conversation_messages,
             model=model,
             max_completion_tokens=max_completion_tokens,
             temperature=temperature,
             response_format=response_format,
-            seed=seed,
             tools=tools,
             tool_choice=tool_choice,
         )
@@ -501,7 +588,7 @@ class GeminiProvider(BaseLLMProvider):
         try:
             response = await client.aio.models.generate_content(
                 model=model,
-                contents=messages,
+                contents=provider_messages,
                 config=GenerateContentConfig(**request_params),
             )
 
@@ -516,19 +603,30 @@ class GeminiProvider(BaseLLMProvider):
                 client=client,
                 response=response,
                 request_params=request_params,
-                messages=messages,
+                messages=provider_messages,
                 tools=tools,
                 tool_dict=tool_dict,
                 response_format=response_format,
                 model=model,
                 post_tool_function=post_tool_function,
-                image_result_keys=image_result_keys,
+                post_response_hook=post_response_hook,
                 tool_handler=tool_handler,
+                return_tool_outputs_only=return_tool_outputs_only,
             )
         except Exception as e:
             traceback.print_exc()
             raise ProviderError(self.get_provider_name(), f"API call failed: {e}", e)
 
+        api_response_id = getattr(response, "id", None)
+        response_id = api_response_id
+        if store:
+            cache_response_id = api_response_id or self.generate_response_id()
+            history_for_cache = self.append_assistant_message_to_history(
+                conversation_messages, content
+            )
+            self.persist_conversation_history(cache_response_id, history_for_cache)
+            response_id = cache_response_id
+
         # Calculate cost
         cost = CostCalculator.calculate_cost(
             model, input_toks, output_toks, cached_toks
@@ -544,4 +642,5 @@ class GeminiProvider(BaseLLMProvider):
             output_tokens_details=output_details,
             cost_in_cents=cost,
             tool_outputs=tool_outputs,
+            response_id=response_id,
         )
