@@ -29,14 +29,6 @@ class MistralProvider(BaseLLMProvider):
     def get_provider_name(self) -> str:
         return "mistral"
 
-    def supports_tools(self, model: str) -> bool:
-        # Most Mistral models support function calling
-        return True
-
-    def supports_response_format(self, model: str) -> bool:
-        # Mistral supports structured outputs via response_format
-        return True
-
     def _get_media_type(self, img_data: str) -> str:
         """Detect media type from base64 image data."""
         try:
@@ -110,7 +102,6 @@ class MistralProvider(BaseLLMProvider):
         max_completion_tokens: Optional[int] = None,
         temperature: float = 0.0,
         response_format=None,
-        seed: int = 0,
         tools: Optional[List[Callable]] = None,
         tool_choice: Optional[str] = None,
         store: bool = True,
@@ -118,7 +109,6 @@ class MistralProvider(BaseLLMProvider):
         timeout: int = 600,
         prediction: Optional[Dict[str, str]] = None,
         reasoning_effort: Optional[str] = None,
-        mcp_servers: Optional[List[Dict[str, Any]]] = None,
         **kwargs,
     ) -> Tuple[Dict[str, Any], List[Dict[str, str]]]:
         """Create the parameter dict for Mistral's chat.complete()."""
@@ -134,10 +124,6 @@ class MistralProvider(BaseLLMProvider):
         if max_completion_tokens is not None:
             params["max_tokens"] = max_completion_tokens
 
-        # Handle random seed
-        if seed != 0:
-            params["random_seed"] = seed
-
         # Handle structured output
         if response_format:
             if isinstance(response_format, type) and hasattr(
@@ -374,7 +360,6 @@ class MistralProvider(BaseLLMProvider):
         max_completion_tokens: Optional[int] = None,
         temperature: float = 0.0,
         response_format=None,
-        seed: int = 0,
         tools: Optional[List[Callable]] = None,
         tool_choice: Optional[str] = None,
         store: bool = True,
@@ -383,7 +368,6 @@ class MistralProvider(BaseLLMProvider):
         prediction: Optional[Dict[str, str]] = None,
         reasoning_effort: Optional[str] = None,
         post_tool_function: Optional[Callable] = None,
-        mcp_servers: Optional[List[Dict[str, Any]]] = None,
         tool_budget: Optional[Dict[str, int]] = None,
         **kwargs,
     ) -> LLMResponse:
@@ -391,7 +375,9 @@ class MistralProvider(BaseLLMProvider):
         from mistralai import Mistral
 
         # Create a ToolHandler instance with tool_budget if provided
-        tool_handler = self.create_tool_handler_with_budget(tool_budget)
+        tool_handler = self.create_tool_handler_with_budget(
+            tool_budget, None, kwargs.get("tool_output_max_tokens")
+        )
 
         if post_tool_function:
             tool_handler.validate_post_tool_function(post_tool_function)
@@ -411,7 +397,6 @@ class MistralProvider(BaseLLMProvider):
             tools=tools,
             tool_choice=tool_choice,
             response_format=response_format,
-            seed=seed,
             timeout=timeout,
         )
 
