@@ -11,6 +11,7 @@ from .providers import (
     TogetherProvider,
     DeepSeekProvider,
     MistralProvider,
+    GrokProvider,
 )
 from .providers.base import LLMResponse
 from .exceptions import LLMError, ConfigurationError
@@ -18,6 +19,7 @@ from .config import LLMConfig
 from .llm_providers import LLMProvider
 from .citations import citations_tool
 from copy import deepcopy
+from .utils_mcp import get_mcp_tools
 
 # Keep the original LLMResponse for backwards compatibility
 # (it's now defined in providers.base but we re-export it here)
@@ -66,6 +68,7 @@ def get_provider_instance(
         "together": TogetherProvider,
         "deepseek": DeepSeekProvider,
         "mistral": MistralProvider,
+        "grok": GrokProvider,
         "alibaba": OpenAIProvider,  # Alibaba uses OpenAI-compatible API
     }
 
@@ -93,7 +96,7 @@ async def chat_async(
     max_completion_tokens: Optional[int] = None,
     temperature: float = 0.0,
     response_format=None,
-    seed: int = 0,
+    verbosity: Optional[str] = None,
     store: bool = True,
     metadata: Optional[Dict[str, str]] = None,
     timeout: int = 600,
@@ -105,11 +108,19 @@ async def chat_async(
     tool_choice: Optional[str] = None,
     max_retries: Optional[int] = None,
     post_tool_function: Optional[Callable] = None,
+    post_response_hook: Optional[Callable] = None,
     config: Optional[LLMConfig] = None,
-    mcp_servers: Optional[List[Dict[str, Any]]] = None,
+    mcp_servers: Optional[List[str]] = None,
     image_result_keys: Optional[List[str]] = None,
     tool_budget: Optional[Dict[str, int]] = None,
     insert_tool_citations: bool = False,
+    citations_instructions: Optional[str] = None,
+    parallel_tool_calls: bool = False,
+    tool_output_max_tokens: int = 10000,
+    previous_response_id: Optional[str] = None,
+    citations_model: Optional[str] = None,
+    citations_provider: Optional[Union[LLMProvider, str]] = None,
+    citations_excluded_tools: Optional[List[str]] = None,
 ) -> LLMResponse:
     """
     Execute a chat completion with explicit provider parameter.
@@ -121,7 +132,6 @@ async def chat_async(
         max_completion_tokens: Maximum tokens to generate
         temperature: Sampling temperature (0.0 to 1.0)
         response_format: Structured output format (Pydantic model)
-        seed: Random seed for reproducibility
         store: Whether to store the conversation
         metadata: Additional metadata
         timeout: Request timeout in seconds
@@ -132,12 +142,19 @@ async def chat_async(
         tools: List of tools the model can call
         tool_choice: Tool calling behavior ("auto", "required", function name)
         max_retries: Maximum number of retry attempts
-        post_tool_function: Function to call after each tool execution
+        post_tool_function: Function to call after each tool execution. Must have parameters: function_name, input_args, tool_result, tool_id
+        post_response_hook: Function to call after each response is received from the model. Must have parameters: response, messages
         config: LLM configuration object
-        mcp_servers: List of MCP server configurations (Anthropic only)
+        mcp_servers: List of MCP server urls for streamable http servers (e.g., ["http://localhost:8000/mcp", "http://localhost:8001/mcp"])
         image_result_keys: List of keys to check in tool results for image data (e.g., ['image_base64', 'screenshot_data'])
         tool_budget: Dictionary mapping tool names to maximum allowed calls. Tools not in the dictionary have unlimited calls.
         insert_tool_citations: If True, adds citations to the response using tool outputs as source documents (OpenAI and Anthropic only)
+        citations_model: Optional model to use specifically for generating citations. If provided, the provider is inferred from the model name unless citations_provider is set. If not provided, the main model is used.
+        citations_provider: Optional provider to use for generating citations. Overrides provider inference from citations_model. Accepts LLMProvider enum or string name. If not provided, the main provider is used.
+        citations_excluded_tools: Optional list of tool names to exclude from the documents sent for citation generation.
+        parallel_tool_calls: Enable parallel tool calls when set to True (default: False)
+        tool_output_max_tokens: Maximum tokens allowed in tool outputs (default: 10000). Set to -1 to disable the check.
+        previous_response_id: Optional id of a previous response when continuing conversations (supported for OpenAI, Anthropic/Grok, Gemini).
 
     Returns:
         LLMResponse object containing the result
@@ -159,6 +176,21 @@ async def chat_async(
     base_delay = 1  # Initial delay in seconds
     error_trace = None
 
+    # validate that mcp servers are simple strings and nothing else
+    if mcp_servers:
+        for mcp_server in mcp_servers:
+            if not isinstance(mcp_server, str):
+                raise ValueError("mcp_servers must be a list of strings")
+
+    if mcp_servers:
+        mcp_tools = []
+        for mcp_server in mcp_servers:
+            mcp_tools.extend(await get_mcp_tools(mcp_server))
+        if tools:
+            tools = tools + mcp_tools
+        else:
+            tools = mcp_tools
+
     for attempt in range(max_retries):
         try:
             # Use backup provider/model on subsequent attempts
@@ -172,20 +204,76 @@ async def chat_async(
                     current_model = backup_model
 
             # Validate provider support for citations before execution
+            # Determine which provider/model will be used for citations if enabled
             if insert_tool_citations:
+                # Determine provider for the main call
                 if isinstance(current_provider, str):
-                    provider_enum = LLMProvider(current_provider.lower())
+                    provider_enum_for_main = LLMProvider(current_provider.lower())
                 else:
-                    provider_enum = current_provider
+                    provider_enum_for_main = current_provider
 
-                if provider_enum not in [LLMProvider.OPENAI, LLMProvider.ANTHROPIC]:
+                # Determine citations provider and model
+                # Provider precedence: explicit citations_provider > inferred from citations_model > main provider
+                if citations_provider is not None:
+                    if isinstance(citations_provider, str):
+                        citations_provider_enum = LLMProvider(
+                            citations_provider.lower()
+                        )
+                    else:
+                        citations_provider_enum = citations_provider
+                elif citations_model:
+                    try:
+                        citations_provider_enum = map_model_to_provider(citations_model)
+                    except Exception as e:
+                        raise ConfigurationError(
+                            f"Unable to map citations_model '{citations_model}' to a provider: {e}"
+                        )
+                else:
+                    citations_provider_enum = provider_enum_for_main
+
+                citations_model_to_use = (
+                    citations_model if citations_model else current_model
+                )
+
+                # If both provider and model were provided, check for an obvious mismatch and warn via error
+                if citations_provider is not None and citations_model:
+                    try:
+                        inferred = map_model_to_provider(citations_model)
+                        if inferred != citations_provider_enum:
+                            raise ConfigurationError(
+                                f"citations_model '{citations_model}' appears to map to provider '{inferred.value}', "
+                                f"but citations_provider was set to '{citations_provider_enum.value}'. Please align them."
+                            )
+                    except Exception:
+                        # If mapping fails, we already raised earlier; keep going otherwise
+                        pass
+
+                # Validate provider support for citations
+                if citations_provider_enum not in [
+                    LLMProvider.OPENAI,
+                    LLMProvider.ANTHROPIC,
+                ]:
                     raise ValueError(
                         "insert_tool_citations is only supported for OpenAI and Anthropic providers"
                     )
 
+                # Validate that we have credentials for the citations provider
+                provider_name = (
+                    citations_provider_enum.value
+                    if hasattr(citations_provider_enum, "value")
+                    else str(citations_provider_enum)
+                )
+                if not config.validate_provider_config(provider_name):
+                    raise ConfigurationError(
+                        f"No API key found for citations provider '{provider_name}'"
+                    )
+
             # Get provider instance
             provider_instance = get_provider_instance(current_provider, config)
 
+            if post_response_hook:
+                provider_instance.validate_post_response_hook(post_response_hook)
+
             # Execute the chat completion
             response = await provider_instance.execute_chat(
                 messages=messages,
@@ -193,7 +281,7 @@ async def chat_async(
                 max_completion_tokens=max_completion_tokens,
                 temperature=temperature,
                 response_format=response_format,
-                seed=seed,
+                verbosity=verbosity,
                 tools=tools,
                 tool_choice=tool_choice,
                 store=store,
@@ -202,33 +290,79 @@ async def chat_async(
                 prediction=prediction,
                 reasoning_effort=reasoning_effort,
                 post_tool_function=post_tool_function,
-                mcp_servers=mcp_servers,
+                post_response_hook=post_response_hook,
                 image_result_keys=image_result_keys,
                 tool_budget=tool_budget,
+                parallel_tool_calls=parallel_tool_calls,
+                tool_output_max_tokens=tool_output_max_tokens,
+                previous_response_id=previous_response_id,
+                return_tool_outputs_only=insert_tool_citations,
             )
 
-            # Process citations if requested and provider supports it
+            # Process citations if requested and we have tool outputs
             if insert_tool_citations and response.tool_outputs:
-                # Get the provider enum (already validated above)
+                # Determine citations provider/model (recalculate to use the actual values)
                 if isinstance(current_provider, str):
-                    provider_enum = LLMProvider(current_provider.lower())
+                    provider_enum_for_main = LLMProvider(current_provider.lower())
                 else:
-                    provider_enum = current_provider
+                    provider_enum_for_main = current_provider
+
+                # Provider precedence: explicit citations_provider > inferred from citations_model > main provider
+                if citations_provider is not None:
+                    if isinstance(citations_provider, str):
+                        citations_provider_enum = LLMProvider(
+                            citations_provider.lower()
+                        )
+                    else:
+                        citations_provider_enum = citations_provider
+                elif citations_model:
+                    citations_provider_enum = map_model_to_provider(citations_model)
+                else:
+                    citations_provider_enum = provider_enum_for_main
+
+                citations_model_to_use = (
+                    citations_model if citations_model else current_model
+                )
 
                 # Get the original user question
                 original_question = get_original_user_question(messages)
 
+                # Optionally filter out tool outputs from citation documents
+                tool_outputs_for_citation = response.tool_outputs
+                if citations_excluded_tools:
+                    excluded_set = set(citations_excluded_tools)
+                    tool_outputs_for_citation = [
+                        o
+                        for o in response.tool_outputs
+                        if o.get("name") not in excluded_set
+                    ]
+
                 # Convert tool outputs to documents
-                documents = convert_tool_outputs_to_documents(response.tool_outputs)
+                documents = convert_tool_outputs_to_documents(tool_outputs_for_citation)
+
+                # If no documents remain after filtering, skip citation enhancement
+                if not documents:
+                    return response
 
                 # Call citations tool to add citations to the response
+                if citations_instructions:
+                    citation_instructions = citations_instructions
+                elif response.content:
+                    citation_instructions = (
+                        "Add citations to the following response using the tool outputs "
+                        f"as source documents: {response.content}"
+                    )
+                else:
+                    citation_instructions = (
+                        "Use the tool outputs as source documents to answer the user's question. "
+                        "Add citations that reference the relevant tool call for each fact."
+                    )
                 citation_blocks = await citations_tool(
                     question=original_question,
-                    instructions="Add citations to the following response using the tool outputs as source documents: "
-                    + response.content,
+                    instructions=citation_instructions,
                     documents=documents,
-                    model=current_model,
-                    provider=provider_enum,
+                    model=citations_model_to_use,
+                    provider=citations_provider_enum,
                     max_tokens=max_completion_tokens or 16000,
                     verbose=False,  # Don't show progress for internal citation processing
                 )
@@ -245,6 +379,13 @@ async def chat_async(
                         response.citations = citation_blocks
                     # If citation processing failed or returned empty content, keep original content
 
+                # Add citation generation cost to the overall cost if available
+                citation_cost = getattr(citation_blocks, "cost_in_cents", None)
+                if citation_cost is not None:
+                    response.cost_in_cents = (
+                        response.cost_in_cents or 0
+                    ) + citation_cost
+
             return response
 
         except Exception as e:
@@ -291,6 +432,8 @@ def map_model_to_provider(model: str) -> LLMProvider:
         return LLMProvider.ANTHROPIC
     elif model.startswith("gemini"):
         return LLMProvider.GEMINI
+    elif model.startswith("grok"):
+        return LLMProvider.GROK
     elif (
         model.startswith("gpt")
         or model.startswith("o1")
